#define INCL_BASE
#define INCL_PM

#include <os2.h>
#include <stdio.h>
#include "..\plugin.h"

long _System FpVersionCheck(long apLevel, long* level)
{
  if (apLevel ==1 ) {
    *level = 1;
    return FP_NORMAL;
  }
  return FP_LEVEL_ERROR;
}

long _System FpGetPlugInInfo(long level, void* data, long* need)
{
  if (level != 1) {
    return FP_LEVEL_ERROR;
  }
  if (data == 0) {
    *need = sizeof(FilePlugInData1);
  } else {
    FilePlugInData1* param = (FilePlugInData1*)data;
    param->readExtensionList = "bmp";
    param->writeExtension = "bmp";
    param->version = "1.0";
    param->description = "QueenCy plugin for bitmap";
    param->copyright = "(c)Masami Kitamura";
    param->comment = "OS/2 and Window bitmap support";
  }
  return FP_NORMAL;
}

long _System FpGetFileInfo(long level, void* param)
{
  FilePlugInParam1* p1 = (FilePlugInParam1*)param;
  HFILE fileHandle = 0;
  ULONG action = 0;
  APIRET rc;
  BITMAPFILEHEADER2 bmpHeader;
  ULONG readLength = 0;

  rc = DosOpen(p1->fileName,
	       &fileHandle,
	       &action,
	       0L,
	       FILE_READONLY,
	       OPEN_ACTION_FAIL_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
	       OPEN_ACCESS_READONLY | OPEN_SHARE_DENYNONE | OPEN_FLAGS_SEQUENTIAL,
	       NULL);
  if (rc != 0) {
    return FP_OPEN_ERROR;
  }

  rc = DosRead(fileHandle,
	       (PVOID)&bmpHeader,
	       sizeof(BITMAPFILEHEADER2),
	       &readLength);

  rc = DosClose(fileHandle);

  if (bmpHeader.usType != BFT_BMAP) {
    return FP_OPEN_ERROR;
  }
  if (bmpHeader.bmp2.cbFix == sizeof(BITMAPINFOHEADER)) {
    BITMAPINFOHEADER* bmpInfo = (BITMAPINFOHEADER*)&(bmpHeader.bmp2);
    p1->width = bmpInfo->cx;
    p1->height = bmpInfo->cy;
    p1->resolution = 120;
  } else {
    BITMAPINFOHEADER2* bmpInfo = &(bmpHeader.bmp2);
    p1->width = bmpInfo->cx;
    p1->height = bmpInfo->cy;
    p1->resolution = bmpInfo->cxResolution * 2.54 / 100;
  }

  return FP_NORMAL;
}

long _System FpRead(long level, void* parm)
{
  FilePlugInParam1* p1 = (FilePlugInParam1*)parm;

  HDIR dirHandle = HDIR_SYSTEM;
  FILEFINDBUF3 findBuf;
  ULONG fileNames = 1;
  APIRET rc;
  ULONG fileLen;
  UCHAR* bmpBuf;
  HFILE fileHandle = 0;
  ULONG action = 0;
  ULONG readLength = 0;
  long planes;
  long bitCount;
  long width;
  long height;
  long headerType;
  long x;
  long y;
  UCHAR* bits;
  PBITMAPFILEHEADER2 pbmpHeader2;
  long wWidth2 ;

  rc = DosFindFirst(p1->fileName,
		    &dirHandle,
		    FILE_NORMAL,
		    (PVOID)&findBuf,
		    sizeof(FILEFINDBUF3),
		    &fileNames,
		    FIL_STANDARD);
  if (rc != 0) {
    return FP_READ_ERROR;
  }
  fileLen = findBuf.cbFile;

  rc = DosFindClose(dirHandle);

  /* rbg}bvt@C̓ǂݍݗp̈mۂB */
  rc = DosAllocMem((PPVOID)&bmpBuf,
		   fileLen,
		   PAG_COMMIT | PAG_WRITE);
  if (rc != 0) {
    return FP_READ_ERROR;
  }

  /* rbg}bvt@C̓ǂݍ݁B */
  /* rbg}bvt@C̃I[v */
  rc = DosOpen(p1->fileName,
	       &fileHandle,
	       &action,
	       0L,
	       FILE_READONLY,
	       OPEN_ACTION_FAIL_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
	       OPEN_ACCESS_READONLY | OPEN_SHARE_DENYNONE | OPEN_FLAGS_SEQUENTIAL,
	       NULL);
  if (rc != 0) {
    rc = DosFreeMem((PVOID)bmpBuf);
    return FP_READ_ERROR;
  }

  /* rbg}bvt@C̓ǂݍ */
  rc = DosRead(fileHandle,
	       (PVOID)bmpBuf,
	       fileLen,
	       &readLength);
  if (rc != 0) {
    rc = DosFreeMem((PVOID)bmpBuf);
    return FP_READ_ERROR;
  }

  /* rbg}bvt@C̃N[Y */
  rc = DosClose(fileHandle);

  /* ǂݍ񂾃rbg}bvt@C̃`FbN */
  pbmpHeader2 = (PBITMAPFILEHEADER2)((PUCHAR)bmpBuf);
  if (pbmpHeader2->usType != BFT_BMAP) {
    rc = DosFreeMem((PVOID)bmpBuf);
    return FP_READ_ERROR;
  }

  if (pbmpHeader2->bmp2.cbFix == sizeof(BITMAPINFOHEADER)) {
    /* BITMAPFILEHEADEȐꍇ */
    PBITMAPFILEHEADER pbmpHeader = (PBITMAPFILEHEADER)pbmpHeader2;
    BITMAPINFOHEADER* bmpInfo = &(pbmpHeader->bmp);
    planes = bmpInfo->cPlanes;
    bitCount = bmpInfo->cBitCount;
    width = bmpInfo->cx;
    height = bmpInfo->cy;
    p1->resolution = 120;
    bits = bmpBuf + ((PBITMAPFILEHEADER)bmpBuf)->offBits;
    headerType = 1;
  } else {
    /* BITMAPFILEHEADER2̏ꍇ */
    BITMAPINFOHEADER2* bmpInfo = &(pbmpHeader2->bmp2);
    planes = bmpInfo->cPlanes;
    bitCount = bmpInfo->cBitCount;
    width = bmpInfo->cx;
    height = bmpInfo->cy;
    p1->resolution = bmpInfo->cxResolution * 2.54 / 100;
    bits = bmpBuf + ((PBITMAPFILEHEADER2)bmpBuf)->offBits;
    headerType = 2;
  }

  wWidth2 = ((24 * width + 31) / 32) * 4;
  if ((planes == 1) && (bitCount == 24)) {
    /* 24bpp : 1v[24rbgJEg̏ꍇ */
    UCHAR* wBits = bits;
    PBYTE bmp = p1->bits;
    for (y = 0; y < height * wWidth2; y++) {
      *bmp++ = *wBits++;
    }
  } else if ((planes == 1) && (bitCount == 8)) {
    /* 8bpp : 1v[8rbgJEg̏ꍇ */
    long wWidth = ((8 * width + 31) / 32) * 4;
    if (headerType == 2) {
      RGB2* colorIndex = (RGB2*)((UCHAR*)(&(((PBITMAPFILEHEADER2)bmpBuf)->bmp2))
				 + ((PBITMAPFILEHEADER2)bmpBuf)->bmp2.cbFix);
      for (y = 0; y < height; y++) {
	UCHAR* wBits = bits + (y * wWidth);
	PBYTE bmp = p1->bits + (y * wWidth2);
	for (x = 0; x < width; x++) {
	  RGB2 rgb2 = *(colorIndex + *wBits++);
	  *bmp++ = rgb2.bBlue;
	  *bmp++ = rgb2.bGreen;
	  *bmp++ = rgb2.bRed;
	}
      }
    } else {
      RGB* colorIndex1 = (RGB*)((UCHAR*)(&(((PBITMAPFILEHEADER)bmpBuf)->bmp))
				+ ((PBITMAPFILEHEADER)bmpBuf)->bmp.cbFix);
      for (y = 0; y < height; y++) {
	UCHAR* wBits = bits + (y * wWidth);
	PBYTE bmp = p1->bits + (y * wWidth2);
	for (x = 0; x < width; x++) {
	  RGB rgb1 = *(colorIndex1 + *wBits++);
	  *bmp++ = rgb1.bBlue;
	  *bmp++ = rgb1.bGreen;
	  *bmp++ = rgb1.bRed;
	}
      }
    }
  } else if ((planes == 1) && (bitCount == 4)) {
    /* 4bpp : 1v[4rbgJEg̏ꍇ */
    long wWidth = ((4 * width + 31) / 32) * 4;
    if (headerType == 2) {
      RGB2* colorIndex = (RGB2*)((UCHAR*)(&(((PBITMAPFILEHEADER2)bmpBuf)->bmp2))
				 + ((PBITMAPFILEHEADER2)bmpBuf)->bmp2.cbFix);
      for (y = 0; y < height; y++) {
	UCHAR* wBits = bits + (y * wWidth);
	PBYTE bmp = p1->bits + (y * wWidth2);
	for (x = 0; x < (width + 1) / 2; x++) {
	  UCHAR rgbIndex = *wBits++;
	  RGB2 rgb2 = *(colorIndex + ((rgbIndex >> 4) & 0x0f));
	  *bmp++ = rgb2.bBlue;
	  *bmp++ = rgb2.bGreen;
	  *bmp++ = rgb2.bRed;
	  rgb2 = *(colorIndex + (rgbIndex & 0x0f));
	  *bmp++ = rgb2.bBlue;
	  *bmp++ = rgb2.bGreen;
	  *bmp++ = rgb2.bRed;
	}
      }
    } else {
      RGB* colorIndex1 = (RGB*)((UCHAR*)(&(((PBITMAPFILEHEADER)bmpBuf)->bmp))
				+ ((PBITMAPFILEHEADER)bmpBuf)->bmp.cbFix);
      for (y = 0; y < height; y++) {
	UCHAR* wBits = bits + (y * wWidth);
	PBYTE bmp = p1->bits + (y * wWidth2);
	for (x = 0; x < (width + 1) / 2; x++) {
	  UCHAR rgbIndex = *wBits++;
	  RGB rgb1 = *(colorIndex1 + ((rgbIndex >> 4) & 0x0f));
	  *bmp++ = rgb1.bBlue;
	  *bmp++ = rgb1.bGreen;
	  *bmp++ = rgb1.bRed;
	  rgb1 = *(colorIndex1 + (rgbIndex & 0x0f));
	  *bmp++ = rgb1.bBlue;
	  *bmp++ = rgb1.bGreen;
	  *bmp++ = rgb1.bRed;
	}
      }
    }
  } else {
    rc = DosFreeMem((PVOID)bmpBuf);
    return FP_READ_ERROR;
  }

  rc = DosFreeMem((PVOID)bmpBuf);
  return FP_NORMAL;
}

long _System FpWrite(long level, void* parm)
{
  FilePlugInParam1* p1 = (FilePlugInParam1*)parm;
  long len = ((24 * p1->width + 31) / 32) * 4;
  BITMAPFILEHEADER2 fileHeader = {0};
  HFILE fileHandle = 0;
  ULONG action = 0;
  APIRET rc;
  ULONG writeLength;

  fileHeader.usType = BFT_BMAP;
  fileHeader.xHotspot = 0;
  fileHeader.yHotspot = 0;
  fileHeader.cbSize = sizeof(BITMAPFILEHEADER2);
  fileHeader.offBits = sizeof(BITMAPFILEHEADER2);
  fileHeader.bmp2.cbFix = sizeof(BITMAPINFOHEADER2);
  fileHeader.bmp2.cx = p1->width;
  fileHeader.bmp2.cy = p1->height;
  fileHeader.bmp2.cPlanes = 1;
  fileHeader.bmp2.cBitCount = 24;
  fileHeader.bmp2.ulCompression = BCA_UNCOMP;
  fileHeader.bmp2.cbImage = 0;				/* kȂ̏ꍇ0włBiftHgj */
  fileHeader.bmp2.cxResolution = p1->resolution * 100 / 2.54;
  fileHeader.bmp2.cyResolution = p1->resolution * 100 / 2.54;
  fileHeader.bmp2.usUnits = BRU_METRIC;
  fileHeader.bmp2.usRecording = BRA_BOTTOMUP;
  fileHeader.bmp2.usRendering = BRH_NOTHALFTONED;
  fileHeader.bmp2.ulColorEncoding = BCE_RGB;

  rc = DosOpen(p1->fileName,
	       &fileHandle,
	       &action,
	       0L,
	       FILE_NORMAL,
	       OPEN_ACTION_CREATE_IF_NEW | OPEN_ACTION_REPLACE_IF_EXISTS,
	       OPEN_ACCESS_WRITEONLY | OPEN_SHARE_DENYREADWRITE | OPEN_FLAGS_SEQUENTIAL,
	       NULL);
  if (rc != 0) {
    return FP_WRITE_ERROR;
  }

  rc = DosWrite(fileHandle,
		(PVOID)&fileHeader,
		sizeof(BITMAPFILEHEADER2),
		&writeLength);
  if (rc != 0) {
    return FP_WRITE_ERROR;
  }
  rc = DosWrite(fileHandle,
		(PVOID)(p1->bits),
		len * p1->height,
		&writeLength);
  if (rc != 0) {
    return FP_WRITE_ERROR;
  }
  rc = DosClose(fileHandle);
  return FP_NORMAL;
}

long FpOption(long level, HWND owner, PSZ mainPath)
{
  return FP_NORMAL;
}
